#!/usr/local/bin/perl -w
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
#........................................................................
# This is a pipeline script which calls the pcms modules in the correct
# sequence to make the full set of ssc products.
#........................................................................

  package main;
  use strict;
  use English;
  use Carp;
  use vars qw($testflag $name $version $author $date $currentModule $currentInst $currentStream $currentExpid %moduleStatus %expIds %bit_nums @orderedListOfModules);

  use ModuleResources qw(exception ignore info start success);
  use VerifyODF          qw(numberOfStreams evaluateRules performAction);
  use MakeCIF            qw(numberOfStreams evaluateRules performAction);
  use GlobalHK           qw(numberOfStreams evaluateRules performAction);
  use InstrumentHK       qw(numberOfStreams evaluateRules performAction);
  use ExpChooser         qw(numberOfStreams evaluateRules performAction);
  use USNOExtract        qw(numberOfStreams evaluateRules performAction);
  use MOSEvents          qw(numberOfStreams evaluateRules performAction);
  use pnEvents           qw(numberOfStreams evaluateRules performAction);
  use MakeMOSImage       qw(numberOfStreams evaluateRules performAction);
  use MakePNImage        qw(numberOfStreams evaluateRules performAction);
  use EPICFluctSpec      qw(numberOfStreams evaluateRules performAction);
  use ExpDetectPrep      qw(numberOfStreams evaluateRules performAction);
  use ExpDetect          qw(numberOfStreams evaluateRules performAction);
  use ImageMerge         qw(numberOfStreams evaluateRules performAction);
  use RGSEvents          qw(numberOfStreams evaluateRules performAction);
  use RGSProducts        qw(numberOfStreams evaluateRules performAction);
  use RGSMakeFluxed      qw(numberOfStreams evaluateRules performAction);
  use OMgetFlat          qw(numberOfStreams evaluateRules performAction);
  use OMExpAnalyse       qw(numberOfStreams evaluateRules performAction);
  use OMFastAnalyse      qw(numberOfStreams evaluateRules performAction);
  use OMImageAnalyse     qw(numberOfStreams evaluateRules performAction);
  use OMSourceCombine    qw(numberOfStreams evaluateRules performAction);
  use OMMosaic           qw(numberOfStreams evaluateRules performAction);
  use SrcMerge           qw(numberOfStreams evaluateRules performAction);
#  use OMSrcRect          qw(numberOfStreams evaluateRules performAction);
  use EPICSources        qw(numberOfStreams evaluateRules performAction);
  use EPICSourceProducts qw(numberOfStreams evaluateRules performAction);
  use SendACDS           qw(numberOfStreams evaluateRules performAction);
#  use ACDSDTS            qw(numberOfStreams evaluateRules performAction);
  use ReceiveACDS        qw(numberOfStreams evaluateRules performAction);
  use StampKeywords      qw(numberOfStreams evaluateRules performAction);
  use Finalize           qw(numberOfStreams evaluateRules performAction);

  my $taskname = "HyperMod_4.pl";

  sub makeProds;
  sub runModule;

#  use FindBin;
#  use lib "$FindBin::Bin";
  use local_band_subs_1;
#  use module_subs_0;

  # Prints all output to stderr:
  select(STDERR);

  $testflag = 1;
  $name    = __PACKAGE__;
  $version = '0.4';
  $author  = 'Ian Stewart';
  $date    = '2005-06-17';
#
# ChangeLog
# =========
#
# version 0.4 - 2005-11-21 IMS
# ------------
# + Added Finalize.
#
# version 0.3 - 2005-11-18 IMS
# ------------
# + Added USNOExtract.
#
# version 0.2 - 2005-10-14 IMS
# ------------
# + Can now control which modules have action method called.
#
# version 0.1 - 2005-06-17 IMS
# ------------
# + First sketchy proposal.


  #......................................................................
  # Start:

  @orderedListOfModules = (
    {'name'=>'VerifyODF',	  'doaction'=>0},
    {'name'=>'MakeCIF',		  'doaction'=>0},
    {'name'=>'GlobalHK',	  'doaction'=>0},
    {'name'=>'ExpChooser',	  'doaction'=>0},
    {'name'=>'USNOExtract',	  'doaction'=>0},
    {'name'=>'InstrumentHK',	  'doaction'=>0},
    {'name'=>'MOSEvents',	  'doaction'=>0},
    {'name'=>'pnEvents',	  'doaction'=>0},
    {'name'=>'MakeMOSImage',	  'doaction'=>0},
    {'name'=>'MakePNImage',	  'doaction'=>0},
    {'name'=>'EPICFluctSpec',	  'doaction'=>0},
    {'name'=>'ExpDetectPrep',	  'doaction'=>0},
    {'name'=>'ExpDetect',	  'doaction'=>1},
    {'name'=>'ImageMerge',	  'doaction'=>0},
    {'name'=>'RGSEvents',	  'doaction'=>0},
    {'name'=>'RGSProducts',	  'doaction'=>0},
    {'name'=>'RGSMakeFluxed',	  'doaction'=>0},
    {'name'=>'OMgetFlat',	  'doaction'=>0},
    {'name'=>'OMExpAnalyse',	  'doaction'=>0},
    {'name'=>'OMFastAnalyse',	  'doaction'=>0},
    {'name'=>'OMImageAnalyse',	  'doaction'=>0},
    {'name'=>'OMMosaic',	  'doaction'=>0},
    {'name'=>'OMSourceCombine',	  'doaction'=>0},
#    {'name'=>'OMSrcRect',	  'doaction'=>0},
    {'name'=>'SrcMerge',	  'doaction'=>0},
    {'name'=>'EPICSources',	  'doaction'=>1},
    {'name'=>'EPICSourceProducts','doaction'=>1},
    {'name'=>'SendACDS',	  'doaction'=>0},
    {'name'=>'ReceiveACDS',	  'doaction'=>0},
    {'name'=>'StampKeywords',	  'doaction'=>0},
    {'name'=>'Finalize',	  'doaction'=>0}
  );

### exit if userid is wrong?

### write date and version id to run log
### if version is new, write date, version id and sas versions to version log

### input repository for products, intermediates and logs (env var?)

### finds a list of odf directories

#  %modulesInfo = ();
  %moduleStatus = ();

  %bit_nums = (
    'UNDERSHOOT'         => 30,
    'BAD_E3E4'           => 29,
    'ON_BADROW'          => 28,
    'OUTSIDE_THRESHOLDS' => 26,
    'OUT_OF_CCD_WINDOW'  => 25,
    'ON_BADPIX'          => 21,
    'COSMIC_RAY'         => 19,
    'IN_BAD_FRAME'       => 17,
    'OUT_OF_FOV'         => 16,
    'FLICKERING'         => 15,
    'ON_BADOFFSET'       => 13,
  );

#vvv just for test
my @odf_dirs = ('dummy_value');
#^^^ just for test


  foreach my $odf_dir (@odf_dirs) {
### test it for doability and 'next' if not

print "odf_dir=$odf_dir\n\n";

    if (&makeProds($odf_dir)) {
### call warning;
### write a 'failed' flag (+date, +version id) to odf log;
    } else {
### write 'success' flag (+date, +version id) to odf log;
    }
  }

print "\n   >>> FINISHED! <<<\n";

exit 0;

  #======================================================================
  # Have this as a separate, runable script??? !!!

  #----------------------------------------------------------------------
  sub makeProds {
  #----------------------------------------------------------------------

    my $odf_dir = $_[0];
#    my @orderedListOfModuleNames = qw(VerifyODF MakeCIF GlobalHK
#      ExpChooser InstrumentHK MOSEvents pnEvents MakeMOSImage
#      MakePNImage ExpDetectPrep ExpDetect ImageMerge RGSEvents
#      RGSProducts OMgetFlat OMExpAnalyse OMFastAnalyse OMImageAnalyse
#      OMSourceCombine OMMosaic SrcMerge OMSrcRect EPICSources
#      EPICSourceProducts SendACDS ReceiveACDS StampKeywords
#    );

### get list of insts + exposures (return exception() if fail) and set up stream_num->exposure map. Just bodge it for now:

    %expIds = (
      'epn' => {
        1 => 'S001'
      },
      'emos1' => {
        1 => 'S002'
      },
      'emos2' => {
        1 => 'S003'
      },
      'rgs1' => {
        1 => 'S001'
      },
      'rgs2' => {
        1 => 'S002'
      },
      'om' => {
        1 => 'S401'
      },
    );


#    $currentExpid = $expIds{'epn'}{1};
#print "$currentExpid\n";

#exit 0;

### load common variables so modules know where to find the files, where to write products and what exposures/insts are available.

    $moduleStatus{Initiate}{obs}{1} = 'complete';
    $moduleStatus{ACDSDTS} {obs}{1} = 'complete';
    foreach my $module_ref (@orderedListOfModules) {
      my $modname = $$module_ref{name};
      my $doaction = $$module_ref{doaction};
  return 1 if (&runModule($modname, $doaction));
    }

  return 0;
  }

  #----------------------------------------------------------------------
  sub runModule {
  #----------------------------------------------------------------------

    no strict "refs";

    my ($moduleName, $doaction) = @_;

print "\n#############\nmoduleName=$moduleName\n\n";
    if (!-e "$moduleName.pm") {
      print "ERROR! Can't find $moduleName.pm.\n\n";
  return 2;
    }

    $currentModule = ${$moduleName."::name"};
    my @insts = @{$moduleName."::instList"};
print "About to cycle through insts\n";
    foreach my $inst (@insts) {
print "\n  inst=$inst\n";
      $currentInst = $inst;

      my $numStreams = &{$moduleName."::numberOfStreams"}();
print "  numStreams=$numStreams\n";
      foreach my $stream (1 .. $numStreams) {
print "\n    stream=$stream\n";
        $currentStream = $stream;
        $currentExpid = $expIds{$inst}{$stream};

        &{$moduleName."::evaluateRules"};
        my $status
          = $moduleStatus{$currentModule}{$currentInst}{$currentStream};
print "\n    evaluateRules status=$status\n";

      next if ($status eq 'ignored' || $status eq 'complete');
  return 1 if ($status ne 'queued');

        if ($doaction) {
          &{$moduleName."::performAction"};
        } else {
          $moduleStatus{$currentModule}{$currentInst}{$currentStream} = 'complete';
        }
        $status
          = $moduleStatus{$currentModule}{$currentInst}{$currentStream};

  return 1 if ($status eq 'exception');
      }
    }

  return 0;
  }

  #---------------------------------------------------------------------
  sub template {
  #---------------------------------------------------------------------

  }

  #======================================================================

