! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: l_plot_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some pgplot wrappers.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module l_plot_utils

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling

  ! ssclib modules:
!  use ssc_misc	! for decideOption.
!use temp_ssclib	! for decideOption
use l_ssc_misc_2	! for decideOption.
use l_plot_types

  implicit none 

  private 

  logical ::&
    plotInitialized=.false., plotAutoSizedX=.false., plotAutoSizedY=.false.,&
    plotLogX=.false., plotLogY=.false.,&
    imageAutoScaled=.false.,coloursHaveBeenSet=.false.
  integer ::&
    numPreviousPlots=0,&
    localNumColourPoints = 256,&
    numReservedClrs = 3,&
    minCi=0,maxCi=0,ciLo=0,ciHi=0
  character(100) :: xLabel='', yLabel='', plotLabel=''
  real ::&
    imageMaxVal=0.0, imageMinVal=0.0,&
    worldXLo=0.0, worldXHi=1.0, worldYLo=0.0, worldYHi=1.0,&
    worldMinDataX=0.0, worldMaxDataX=1.0, worldMinDataY=0.0, worldMaxDataY=1.0

  integer, parameter, public ::&
    P_POINT=-1, P_TRIANGLE=7, P_SQUARE=0, P_CIRCLE=4, P_CROSS=2,&
    P_DIAMOND=11, P_XCROSS=5, P_ASTERIX=3, P_FILLEDTRIANGLE=13,&
    P_FILLEDSQUARE=16, P_FILLEDCIRCLE=17,&
    L_SOLID=1, L_DASHED=2, L_DOTTED=4,&
    C_WHITE=1, C_RED=2, C_BLUE=3, C_GREEN=4
  real, parameter ::&
    defaultVpXLo=0.1, defaultVpXHi=0.99, defaultVpYLo=0.1, defaultVpYHi=0.9,&
    xPadFrac = 0.05, yPadFrac = 0.05

  public plotInit
  public setXLog, setYLog
  public plotAutoSizeX, plotAutoSizeY
  public drawPointPlot
  public drawLinePlot
  public drawHistogram
  public finishPlot
  public imageAutoScale
  public drawImage
  public finishImage

  interface drawPointPlot
    module procedure drawPointPlotXY, drawPointPlotY
  end interface
  interface drawLinePlot
    module procedure drawLinePlotXY, drawLinePlotY
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine plotInit(pgplotDevice, xLabelArg, yLabelArg, plotLabelArg)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in)           :: pgplotDevice
    character(*), intent(in), optional :: xLabelArg, yLabelArg, plotLabelArg

    integer :: deviceId, pgopen

    !-------------------------------------------------------------------
    ! Start:

    if (plotInitialized)&
call error('pgplotAlreadyInitialized',"")

    deviceId = pgopen(trim(pgplotDevice))
if (deviceId <= 0) call error('badPgplotDevice',"")

    call pgsvp(defaultVpXLo, defaultVpXHi, defaultVpYLo, defaultVpYHi)
    plotInitialized = .true.

    if (present(xLabelArg))&
      xLabel = xLabelArg
    if (present(yLabelArg))&
      yLabel = yLabelArg
    if (present(plotLabelArg))&
      plotLabel = plotLabelArg

    call pgqcir(minCi, maxCi)
    ciLo = minCi + numReservedClrs
    ciHi = maxCi
    call pgscir(ciLo, ciHi)

  end subroutine plotInit

  !---------------------------------------------------------------------
  subroutine setXLog
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    plotLogX = .true.

  end subroutine setXLog

  !---------------------------------------------------------------------
  subroutine setYLog
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    plotLogY = .true.

  end subroutine setYLog

  !---------------------------------------------------------------------
  subroutine plotAutoSizeX(xVals, xErrs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:)
    real(single), intent(in), optional :: xErrs(size(xVals))

    real :: newXLo, newXHi
!, xPadSize
    logical :: xRangesSet
    logical, allocatable :: xIsOk(:)

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized)&
call error('pgplotNotInitialized',"")

    allocate(xIsOk(size(xVals)))
    if (plotLogX) then
      xIsOk = (xVals > 0.0)
    else
      xIsOk = .true.
    end if

    xRangesSet = .false.
    if (count(xIsOk)>0) then
      if (present(xErrs)) then
        if (plotLogX) then
          newXLo = minval(xVals, xIsOk)
        else
          newXLo = minval(xVals-xErrs, xIsOk)
        end if
        newXHi = maxval(xVals+xErrs, xIsOk)
      else
        newXLo = minval(xVals, xIsOk)
        newXHi = maxval(xVals, xIsOk)
!print *, 'newXLo, newXHi:', newXLo, newXHi
      end if
      xRangesSet = .true.
    end if
!print *, 'xRangesSet?', xRangesSet

    if (plotAutoSizedX) then
!      call pgqwin(xLo, xHi, yLo, yHi)
      if (xRangesSet) then
        if (plotLogX .and. worldMinDataX <= 0.0) then
          worldMinDataX = newXLo
        else
          worldMinDataX = min(worldMinDataX, newXLo)
        end if
        worldMaxDataX = max(worldMaxDataX, newXHi)
      end if

      if (plotLogX) then
        worldMinDataX = log10(worldMinDataX)
        worldMaxDataX = log10(worldMaxDataX)
      end if

    else
      if (xRangesSet) then
        if (plotLogX) then
          worldMinDataX = log10(newXLo)
          worldMaxDataX = log10(newXHi)
        else
          worldMinDataX = newXLo
          worldMaxDataX = newXHi
!print *, 'worldMinDataX, worldMaxDataX:', worldMinDataX, worldMaxDataX
        end if
      else
        worldMinDataX = 0.0
        worldMaxDataX = 1.0
      end if
    end if

if (worldMinDataX >= worldMaxDataX) call error('badOrderXBounds',"")

!    xPadSize = (worldMaxDataX - worldMinDataX) * xPadFrac / (1.0d0 - 2.0d0 * xPadFrac)

!    worldMinDataX = worldMinDataX - xPadSize
!    worldMaxDataX = worldMaxDataX + xPadSize
!print *, 'worldMinDataX, worldMaxDataX after pad applied:', worldMinDataX, worldMaxDataX

!    call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)

    plotAutoSizedX = .true.

  end subroutine plotAutoSizeX

  !---------------------------------------------------------------------
  subroutine plotAutoSizeY(yVals, yErrs)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    real(single), intent(in), optional :: yErrs(size(yVals))

    real :: newYLo, newYHi
!, yPadSize
    logical :: yRangesSet
    logical, allocatable :: yIsOk(:)
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa plotAutoSizeY'

    if (.not.plotInitialized)&
call error('pgplotNotInitialized',"")

    allocate(yIsOk(size(yVals)))
    if (plotLogY) then
      yIsOk = (yVals > 0.0)
    else
      yIsOk = .true.
    end if

    yRangesSet = .false.
    if (count(yIsOk)>0) then
      if (present(yErrs)) then
        if (plotLogY) then
          newYLo = minval(yVals, yIsOk)
        else
          newYLo = minval(yVals-yErrs, yIsOk)
        end if
        newYHi = maxval(yVals+yErrs, yIsOk)
      else
        newYLo = minval(yVals, yIsOk)
        newYHi = maxval(yVals, yIsOk)
if (testFlag) print *, '  bbb - newYLo, newYHi:', newYLo, newYHi
      end if
      yRangesSet = .true.
    end if

    if (plotAutoSizedY) then
!      call pgqwin(yLo, yHi, yLo, yHi)
      if (yRangesSet) then
        if (plotLogY .and. worldMinDataY <= 0.0) then
          worldMinDataY = newYLo
        else
          worldMinDataY = min(worldMinDataY, newYLo)
        end if
        worldMaxDataY = max(worldMaxDataY, newYHi)
if (testFlag) print *, '  ccc - 2nd run - worldMinDataY, worldMaxDataY:', worldMinDataY, worldMaxDataY
      end if

      if (plotLogY) then
        worldMinDataY = log10(worldMinDataY)
        worldMaxDataY = log10(worldMaxDataY)
      end if

    else
      if (yRangesSet) then
        if (plotLogY) then
          worldMinDataY = log10(newYLo)
          worldMaxDataY = log10(newYHi)
        else
          worldMinDataY = newYLo
          worldMaxDataY = newYHi
if (testFlag) print *, '  ccc - 1st run - worldMinDataY, worldMaxDataY:', worldMinDataY, worldMaxDataY
        end if
      else
        worldMinDataY = 0.0
        worldMaxDataY = 1.0
      end if
    end if

if (worldMinDataY >= worldMaxDataY) call error('badOrderYBounds',"")

!    yPadSize = (worldMaxDataY - worldMinDataY) * yPadFrac / (1.0d0 - 2.0d0 * yPadFrac)

!    worldMinDataY = worldMinDataY - yPadSize
!    worldMaxDataY = worldMaxDataY + yPadSize

!    call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)

    plotAutoSizedY = .true.
if (testFlag) print *, '  zzz plotAutoSizeY'

  end subroutine plotAutoSizeY

  !---------------------------------------------------------------------
  subroutine drawPointPlotY(yVals, pointStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    integer,      intent(in), optional :: pointStyle
    logical,      intent(in), optional :: finished

    real(single), allocatable :: xVals(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    allocate(xVals(size(yVals)))
    xVals = (/(real(i-1, single), i=1,size(yVals))/)
    call drawPointPlotXY(xVals, yVals, pointStyle=pointStyle, finished=finished)

  end subroutine drawPointPlotY

  !---------------------------------------------------------------------
!  subroutine drawPointPlot(xVals, yVals, xErrs, yErrs, pointStyle, finished)
  subroutine drawPointPlotXY(xVals, yVals, yErrs, pointStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
!    real(single), intent(in), optional :: xErrs(size(xVals)), yErrs(size(xVals))
    real(single), intent(in), optional :: yErrs(size(xVals))
    integer,      intent(in), optional :: pointStyle
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: localPointStyle, numOkVals
!    logical, allocatable :: xIsOk(:), yIsOk(:), valIsOk
    real, pointer ::&
      localXVals(:), localYVals(:), localYBots(:), localYTops(:)
!integer :: i

    !-------------------------------------------------------------------
    ! Start:

!    if (.not.plotAutoSized) call plotAutoSize(xVals, yVals, xErrs, yErrs)
!    if (.not.plotAutoSized) call plotAutoSizeXY(xVals, yVals, yErrs=yErrs)
if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
  if (.not.plotAutoSizedX) call plotAutoSizeX(xVals)
  if (.not.plotAutoSizedY) call plotAutoSizeY(yVals, yErrs)
!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
  call setWorldRanges
end if
    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.
    localPointStyle = decideOption(pointStyle, P_POINT)

    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

    call getLocalVals(xVals, yVals, yErrs, localXVals, localYVals&
      , localYBots, localYTops)
    numOkVals = size(localXVals)
!do i = 1, numOkVals
!  print *, i, localXVals(i), localYBots(i), localYTops(i)
!end do

    if (numOkVals > 0) then
      call pgpt(numOkVals, localXVals, localYVals, localPointStyle)
      if (present(yErrs))&
        call pgerry(numOkVals, localXVals, localYBots, localYTops, 1.0)
    end if

    if (localFinished) call finishPlot
    deallocate(localXVals, localYVals, localYBots, localYTops)

  end subroutine drawPointPlotXY

  !---------------------------------------------------------------------
  subroutine drawLinePlotY(yVals, lineStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: yVals(:)
    integer,      intent(in), optional :: lineStyle
    logical,      intent(in), optional :: finished

    real(single), allocatable :: xVals(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    allocate(xVals(size(yVals)))
    xVals = (/(real(i-1, single), i=1,size(yVals))/)
    call drawLinePlotXY(xVals, yVals, lineStyle, finished)

  end subroutine drawLinePlotY

  !---------------------------------------------------------------------
  subroutine drawLinePlotXY(xVals, yVals, lineStyle, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
    integer,      intent(in), optional :: lineStyle
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: localLineStyle, oldLineStyle, numOkVals
    real, pointer ::&
      localXVals(:), localYVals(:)

logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa drawLinePlot'

!    if (.not.plotAutoSized) call plotAutoSizeXY(xVals, yVals)
!print *, 'plotAutoSizedX?', plotAutoSizedX
!print *, 'plotAutoSizedY?', plotAutoSizedY
if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
!print *, minval(xVals), maxval(xVals)
  if (.not.plotAutoSizedX) call plotAutoSizeX(xVals)
  if (.not.plotAutoSizedY) call plotAutoSizeY(yVals)
!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
  call setWorldRanges
end if
if (testFlag) print *, '  bbb drawLinePlot'
    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.
    localLineStyle = decideOption(lineStyle, L_SOLID)

if (testFlag) print *, '  ccc drawLinePlot'
    call pgqls(oldLineStyle)
if (testFlag) print *, '  ddd drawLinePlot'
    call pgsls(localLineStyle)

if (testFlag) print *, '  eee drawLinePlot'
    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

if (testFlag) print *, '  fff drawLinePlot'
    call getLocalVals(xVals, yVals, localXVals=localXVals&
      , localYVals=localYVals)

    numOkVals = size(localXVals)
if (testFlag) then
  print *, '  ggg drawLinePlot'
  print *, '  ggg0 - numOkVals:', numOkVals
  print *, '  ggg1 - xVals:', xVals
  print *, '  ggg2 - yVals:', yVals
  print *, '  ggg3 - localYVals:', localYVals
  print *, '  ggg4 - localYVals:', localYVals
end if
    if (numOkVals > 0)&
      call pgline(numOkVals, localXVals, localYVals)
!    numPoints = size(xVals)
!    call pgline(numPoints, xVals, yVals)

if (testFlag) print *, '  hhh drawLinePlot'
    call pgsls(oldLineStyle)
if (testFlag) print *, '  iii drawLinePlot'
    if (localFinished) call finishPlot
if (testFlag) print *, '  jjj drawLinePlot'
    deallocate(localXVals, localYVals)
if (testFlag) print *, '  zzz drawLinePlot'

  end subroutine drawLinePlotXY

  !---------------------------------------------------------------------
  subroutine drawHistogram(binEdges, binVals, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: binVals(:)
    real(single), intent(in)           :: binEdges(size(binVals)+1)
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: numBins, i
    real :: someXTooLowToSeeOnPlot, someYTooLowToSeeOnPlot, baseY
    real, allocatable :: xs(:), ys(:)

    !-------------------------------------------------------------------
    ! Start:

!    if (.not.plotAutoSized) call plotAutoSize(binEdges, binVals)
if (.not.(plotAutoSizedX .and. plotAutoSizedY)) then
  if (.not.plotAutoSizedX) call plotAutoSizeX(binEdges)
  if (.not.plotAutoSizedY) call plotAutoSizeY(binVals)
!  call pgswin(worldMinDataX, worldMaxDataX, worldMinDataY, worldMaxDataY)
  call setWorldRanges
end if
    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.

    numBins = size(binVals)
if (any(binEdges(1:numBins)>binEdges(2:))) call error('badOrderBinEdges',"")

    call pgsci(numPreviousPlots+1)
    numPreviousPlots = numPreviousPlots + 1

    allocate(xs(numBins+1), ys(numBins))

    if (plotLogX) then
      someXTooLowToSeeOnPlot = getTooLowX()
      where(binEdges <= 0.0)
        xs = someXTooLowToSeeOnPlot
      elsewhere
        xs = log10(binEdges)
      end where
    else
      xs = binEdges
    end if

    if (plotLogY) then
      someYTooLowToSeeOnPlot = getTooLowY()
      where(binVals <= 0.0)
        ys = someYTooLowToSeeOnPlot
      elsewhere
        ys = log10(binVals)
      end where
      baseY = someYTooLowToSeeOnPlot
    else
      ys = binVals
      baseY = 0.0
    end if

    call pgmove(xs(1), baseY)
    do i = 1, numBins
      call pgdraw(xs(i), ys(i))
      call pgmove(xs(i), ys(i))
      call pgdraw(xs(i+1), ys(i))
      call pgmove(xs(i+1), ys(i))
    end do
    call pgdraw(xs(numBins+1),baseY)

    if (localFinished) call finishPlot

  end subroutine drawHistogram

  !---------------------------------------------------------------------
  subroutine getLocalVals(xVals, yVals, yErrs, localXVals, localYVals&
  , localYBots, localYTops)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: xVals(:), yVals(size(xVals))
!    real(single), intent(in), optional :: xErrs(size(xVals)), yErrs(size(xVals))
    real(single), intent(in), optional :: yErrs(size(xVals))
    real,         pointer              :: localXVals(:),&
                                          localYVals(:)
    real,         pointer,    optional :: localYBots(:),&
                                          localYTops(:)

    integer :: numPoints, numOkVals
    logical, allocatable :: xIsOk(:), yIsOk(:), valIsOk(:)
    real, allocatable :: localYErrs(:)
    real :: someYTooLowToSeeOnPlot
integer :: i
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa getLocalVals'

    numPoints = size(xVals)
if (testFlag) then
  do i = 1, numPoints
    print *, '  ', i, xVals(i), yVals(i)
  end do
  if (present(yErrs)) print *, '  present(yErrs) T'
  print *, '  plotLogX', plotLogX
  print *, '  plotLogY', plotLogY
end if

    allocate(xIsOk(numPoints), yIsOk(numPoints), valIsOk(numPoints))
    if (plotLogX) then
      xIsOk = (xVals > 0.0)
    else
      xIsOk = .true.
    end if
    if (plotLogY) then
      yIsOk = (yVals > 0.0)
    else
      yIsOk = .true.
    end if
    valIsOk = (xIsOk .and. yIsOk)

    numOkVals = count(valIsOk)
if (testFlag) print *, '  numOkVals', numOkVals
    allocate(localXVals(numOkVals), localYVals(numOkVals))
    localXVals = pack(xVals, valIsOk)
    localYVals = pack(yVals, valIsOk)
if (testFlag) print *, '  bbb getLocalVals'

    if (present(yErrs)) then
if (.not.present(localYBots) .or. .not.present(localYTops)) call error('',"")
      allocate(localYErrs(numOkVals))
      allocate(localYBots(numOkVals), localYTops(numOkVals))
      localYErrs = pack(yErrs, valIsOk)
      localYBots = localYVals - localYErrs
      localYTops = localYVals + localYErrs
!do i = 1, numOkVals
!  print *, i, localXVals(i), localYErrs(i), localYBots(i), localYTops(i)
!end do
      if (plotLogY) then
        localYTops = log10(localYTops)
        someYTooLowToSeeOnPlot = getTooLowY()
        where(localYBots>0.0)
          localYBots = log10(localYBots)
        elsewhere
          localYBots = someYTooLowToSeeOnPlot
        end where 
      end if
if (testFlag) then
  do i = 1, numOkVals
    print *, i, localYBots(i), localYTops(i)
  end do
end if
    end if
if (testFlag) print *, '  ccc getLocalVals'

    if (plotLogX) localXVals = log10(localXVals)
    if (plotLogY) localYVals = log10(localYVals)

if (testFlag) print *, '  zzz getLocalVals'

  end subroutine getLocalVals

  !---------------------------------------------------------------------
  subroutine finishPlot
  !---------------------------------------------------------------------

    implicit none

    character(10) :: xOpt, yOpt

    !-------------------------------------------------------------------
    ! Start:

    if (plotLogX) then
      xOpt = 'BCNTSL'
    else
      xOpt = 'BCNTS'
    end if
    if (plotLogY) then
      yOpt = 'BCNTSL'
    else
      yOpt = 'BCNTS'
    end if
    call pgsci(1)
    call pgsfs(2)
    call pgbox(xOpt, 0.0, 0, yOpt, 0.0, 0)
    call pglab(xLabel, yLabel, plotLabel)
    call pgend
    plotInitialized = .false.

  end subroutine finishPlot

  !---------------------------------------------------------------------
  real function getTooLowX()
  !---------------------------------------------------------------------

    implicit none

    real :: minx, maxx, miny, maxy

    !-------------------------------------------------------------------
    ! Start:

    call pgqwin(minx, maxx, miny, maxy)
    getTooLowX = minx - (maxx - minx)

  end function getTooLowX

  !---------------------------------------------------------------------
  real function getTooLowY()
  !---------------------------------------------------------------------

    implicit none

    real :: minx, maxx, miny, maxy

    !-------------------------------------------------------------------
    ! Start:

    call pgqwin(minx, maxx, miny, maxy)
    getTooLowY = miny - (maxy - miny)

  end function getTooLowY

  !---------------------------------------------------------------------
  subroutine imageAutoScale(array)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: array(:,:)

    integer :: xiSize, yiSize
    real :: xLo, xHi, yLo, yHi

    !-------------------------------------------------------------------
    ! Start:

    if (.not.plotInitialized)&
call error('pgplotNotInitialized',"")

    imageMinVal = minval(array)
    imageMaxVal = maxval(array)

    xiSize = size(array,1)
    yiSize = size(array,2)

    xLo = 0.5
    xHi = xiSize + 0.5
    yLo = 0.5
    yHi = yiSize + 0.5

!    call pgswin(xLo, xHi, yLo, yHi)
    call pgwnad(xLo, xHi, yLo, yHi)

    imageAutoScaled = .true.

  end subroutine imageAutoScale

  !---------------------------------------------------------------------
  subroutine drawImage(array, finished)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)           :: array(:,:)
    logical,      intent(in), optional :: finished

    logical :: localFinished
    integer :: xiSize, yiSize

    !-------------------------------------------------------------------
    ! Start:

    if (.not.imageAutoScaled) call imageAutoScale(array)
    localFinished = decideOption(finished, .true.) ! in ssclib/ssc_misc.

    xiSize = size(array,1)
    yiSize = size(array,2)

    if (coloursHaveBeenSet) then
      call pgimag(real(array), xiSize, yiSize, 1, xiSize, 1, yiSize&
        , imageMaxVal, imageMinVal, pgTrMatrix)
    else
      call pggray(real(array), xiSize, yiSize, 1, xiSize, 1, yiSize&
        , imageMaxVal, imageMinVal, pgTrMatrix)
    end if

    if (localFinished) call finishImage

  end subroutine drawImage

  !---------------------------------------------------------------------
  subroutine finishImage
  !---------------------------------------------------------------------

    implicit none

    character(10) :: xOpt, yOpt

    !-------------------------------------------------------------------
    ! Start:

    xOpt = 'BC'
    yOpt = 'BC'

    call pgsci(1)
    call pgsfs(2)
    call pgbox(xOpt, 0.0, 0, yOpt, 0.0, 0)
    call pglab(xLabel, yLabel,'')
    call pgend
    plotInitialized = .false.

  end subroutine finishImage

  !---------------------------------------------------------------------
  subroutine setWorldRanges
  !---------------------------------------------------------------------

    implicit none

    real :: xPadSize, yPadSize

    !-------------------------------------------------------------------
    ! Start:

if (worldMinDataX >= worldMaxDataX) call error('badOrderXBounds',"")
if (worldMinDataY >= worldMaxDataY) call error('badOrderYBounds',"")

    xPadSize = (worldMaxDataX - worldMinDataX) * xPadFrac / (1.0d0 - 2.0d0 * xPadFrac)
    worldXLo = worldMinDataX - xPadSize
    worldXHi = worldMaxDataX + xPadSize

    yPadSize = (worldMaxDataY - worldMinDataY) * yPadFrac / (1.0d0 - 2.0d0 * yPadFrac)
    worldYLo = worldMinDataY - yPadSize
    worldYHi = worldMaxDataY + yPadSize

    call pgswin(worldXLo, worldXHi, worldYLo, worldYHi)

  end subroutine setWorldRanges

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module l_plot_utils
