! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: l_ssc_misc_2
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: l_ssc_misc_2.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module l_ssc_misc_2

  use dal
  use errorHandling

  use constants	! PiOn2
  use array_utils
  use sort_mod
!use coordinate
use dump

  implicit none 

  private 

! ssc_misc:
  public decideOption
  public elementIsInListChar

  interface decideOption
    module procedure decideOptionInt8, decideOptionInt16, decideOptionInt32
    module procedure decideOptionSingle, decideOptionDouble
    module procedure decideOptionLogical, decideOptionChar
  end interface

  contains

  !---------------------------------------------------------------------
  function decideOptionInt8(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    integer(int8), intent(in), optional :: optArg
    integer(int8), intent(in)           :: default
    integer(int8)                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionInt8

  !---------------------------------------------------------------------
  function decideOptionInt16(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in), optional :: optArg
    integer(int16), intent(in)           :: default
    integer(int16)                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionInt16

  !---------------------------------------------------------------------
  function decideOptionInt32(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in), optional :: optArg
    integer(int32), intent(in)           :: default
    integer(int32)                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionInt32

  !---------------------------------------------------------------------
  function decideOptionSingle(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in), optional :: optArg
    real(single), intent(in)           :: default
    real(single)                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionSingle

  !---------------------------------------------------------------------
  function decideOptionDouble(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in), optional :: optArg
    real(double), intent(in)           :: default
    real(double)                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionDouble

  !---------------------------------------------------------------------
  function decideOptionLogical(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in), optional :: optArg
    logical, intent(in)           :: default
    logical                       :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionLogical

  !---------------------------------------------------------------------
  function decideOptionChar(optArg, default) result(result)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in), optional :: optArg
    character(*), intent(in)           :: default
    character(len(default))            :: result

    !-------------------------------------------------------------------
    ! Start

    if (present(optArg)) then
      result = optArg
    else
      result = default
    end if

  end function decideOptionChar

  !---------------------------------------------------------------------
  logical function elementIsInListChar(str, strList)
  !---------------------------------------------------------------------
!*** put this in ssc_misc??

    implicit none

    character(*), intent(in) :: str, strList(:)

    integer :: numElements, i

    !-------------------------------------------------------------------
    ! Start

    elementIsInListChar = .false. ! default
    numElements = size(strList)
    do i = 1, numElements
      if (str .eq. strList(i)) then
        elementIsInListChar = .true.
  return
      end if
    end do

  end function elementIsInListChar

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_ssc_misc_2
